# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Imports -----------------------------------------------------------------

import datetime
import os
import sys
import re
import shutil
import semantic_version
import urllib.request

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

docssrc_dir = os.path.dirname(os.path.abspath(__file__))
project_dir = os.path.dirname(docssrc_dir)

# When building on ReadTheDocs, we can't provide a local version of the Cython
# extensions, so we have to install the latest public version, and avoid
# patching the PYTHONPATH with the local development folder
if os.getenv("READTHEDOCS", "False") != "True":
    sys.path.insert(0, project_dir)

# Download the *See Also* cards from a centralized location so it can be kept
# up-to-date across all projects
with urllib.request.urlopen("https://gist.githubusercontent.com/althonos/5d6bf5a512d64dc951c42a91d5fc3fb3/raw/related.rst") as src:
    with open("related.rst", "wb") as dst:
        shutil.copyfileobj(src, dst)

# -- Project information -----------------------------------------------------

import pyaragorn

# extract the project metadata from the module itself
project = pyaragorn.__name__
author = re.match('(.*) <.*>', pyaragorn.__author__).group(1)
year = datetime.date.today().year
copyright = '{}, {}'.format("2025" if year==2025 else "2025-{}".format(year), author)

# extract the semantic version
semver = semantic_version.Version.coerce(pyaragorn.__version__)
version = str(semver.truncate(level="patch"))
release = str(semver)

# patch the docstring of `pyaragorn` so that we don't show the link to redirect
# to the docs (we don't want to see it when reading the docs already, duh!)
doc_lines = pyaragorn.__doc__.splitlines()
if "See Also:" in doc_lines:
    see_also = doc_lines.index("See Also:")
    pyaragorn.__doc__ = "\n".join(doc_lines[:see_also])

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    # "sphinx.ext.imgconverter",
    "sphinx.ext.napoleon",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.todo",
    "sphinx.ext.extlinks",
    "sphinx_design",
    "sphinxcontrib.jquery",
    "nbsphinx",
    "recommonmark",
    "IPython.sphinxext.ipython_console_highlighting",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    '_build',
    'Thumbs.db',
    '.DS_Store',
    '**.ipynb_checkpoints',
    'requirements.txt'
]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "monokailight"

# The name of the default role for inline references
default_role = "py:obj"

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'pydata_sphinx_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static/js', '_static/bibtex', '_static/css', '_static/json']
html_js_files = ["custom-icon.js"]
html_css_files = ["custom.css"]

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "external_links": [],
    "show_toc_level": 2,
    "use_edit_page_button": True,
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/althonos/pyaragorn",
            "icon": "fa-brands fa-github",
        },
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/pyaragorn",
            "icon": "fa-custom fa-pypi",
        },
    ],
    "logo": {
        "text": "PyARAGORN",
        "image_light": "_images/logo.png",
        "image_dark": "_images/logo.png",
    },
    "navbar_start": ["navbar-logo", "version-switcher"],
    "navbar_align": "left",
    "footer_start": ["copyright"],
    "footer_center": ["sphinx-version"],
    "switcher": {
        "json_url": "https://pyaragorn.readthedocs.io/en/latest/_static/switcher.json",
        "version_match": version,
    }
}

html_context = {
    "github_user": "althonos",
    "github_repo": "pyaragorn",
    "github_version": "main",
    "doc_path": "docs",
}

html_favicon = '_images/favicon.ico'

# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = pyaragorn.__name__


# -- Extension configuration -------------------------------------------------

# -- Options for imgmath extension -------------------------------------------

imgmath_image_format = "svg"

# -- Options for napoleon extension ------------------------------------------

napoleon_include_init_with_doc = True
napoleon_include_special_with_doc = True
napoleon_include_private_with_doc = True
napoleon_use_admonition_for_examples = True
napoleon_use_admonition_for_notes = True
napoleon_use_admonition_for_references = True
napoleon_use_rtype = False

# -- Options for autodoc extension -------------------------------------------

autoclass_content = "class"
autodoc_member_order = 'groupwise'
autosummary_generate = []

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
}

# -- Options for recommonmark extension --------------------------------------

source_suffix = {
    '.rst': 'restructuredtext',
    '.txt': 'markdown',
    '.md': 'markdown',
}

# -- Options for nbsphinx extension ------------------------------------------

nbsphinx_execute = 'auto'
nbsphinx_execute_arguments = [
    "--InlineBackend.figure_formats={'svg', 'pdf'}",
    "--InlineBackend.rc={'figure.dpi': 96}",
]

# -- Options for extlinks extension ------------------------------------------

extlinks = {
    'doi': ('https://doi.org/%s', 'doi:%s'),
    'pmid': ('https://pubmed.ncbi.nlm.nih.gov/%s', 'PMID:%s'),
    'pmc': ('https://www.ncbi.nlm.nih.gov/pmc/articles/PMC%s', 'PMC%s'),
    'isbn': ('https://www.worldcat.org/isbn/%s', 'ISBN:%s'),
    'wiki': ('https://en.wikipedia.org/wiki/%s', '%s'),
}



import itertools
import re
import unittest

from .. import RNAFinder, TMRNAGene, TRNAGene
from . import data


_TRNA_RX = re.compile(r"^(\d+)\s+tRNA-([A-Za-z]{3})\s+(c?)\[(\d+),(\d+)\]\s+([\d.]+)\s+(\d+)\s+\(([a-z]{2,4})\)")
_TMRNA_RX = re.compile(r"^(\d+)\s+tmRNA\s+(c?)\[(\d+),(\d+)]\s+([\d.]+)\s+(\d+),(\d+)\s+([A-Z\*]+)")

def batched(iterable, n, *, strict=False):
    # batched('ABCDEFG', 3) → ABC DEF G
    if n < 1:
        raise ValueError('n must be at least one')
    iterator = iter(iterable)
    for batch in iter(lambda: tuple(itertools.islice(iterator, n)), ()):
        if strict and len(batch) != n:
            raise ValueError('batched(): incomplete batch')
        yield batch

class TestRNAFinder(unittest.TestCase):
    
    def test_default(self):
        record = data.load_record("CP001621.fna.gz")
        lines = data.load_text("CP001621.default.txt").splitlines()
        
        finder = RNAFinder(translation_table=11)
        genes = finder.find_rna(str(record.seq))

        for gene, expected in itertools.zip_longest(genes, batched(lines[2:], 3)):
            self.assertIsNotNone(gene)
            self.assertIsNotNone(expected)
            result, seq, ss = expected
            if gene.type == "tRNA":
                matched = _TRNA_RX.match(result)
                _, aa, complement, begin, end, energy, offset, anticodon = matched.groups()
                self.assertEqual(gene.amino_acid, aa)
                self.assertEqual(gene.begin, int(begin))
                self.assertEqual(gene.end, int(end))
                self.assertEqual(gene.anticodon_offset, int(offset))
                self.assertEqual(gene.anticodon_length, len(anticodon))
                self.assertEqual(gene.anticodon, anticodon)
                self.assertEqual(gene.strand, -1 if complement == "c" else +1)
                self.assertAlmostEqual(gene.energy, float(energy), places=1)
                self.assertEqual(gene.sequence().lower(), seq)
            elif gene.type == "tmRNA":
                matched = _TMRNA_RX.match(result)
                _, complement, begin, end, energy, orf_start, orf_end, peptide = matched.groups()
                self.assertEqual(gene.begin, int(begin))
                self.assertEqual(gene.end, int(end))
                self.assertEqual(gene.orf_offset, int(orf_start))
                self.assertEqual(gene.orf_offset + gene.orf_length, int(orf_end))
                self.assertEqual(gene.peptide(), peptide)
                self.assertEqual(gene.strand, -1 if complement == "c" else +1)
                self.assertAlmostEqual(gene.energy, float(energy), places=1)
                # self.assertEqual(gene.sequence().lower(), seq) # TODO

    def test_trna(self):
        record = data.load_record("CP001621.fna.gz")
        finder = RNAFinder(translation_table=11, tmrna=False, trna=True)
        for gene in finder.find_rna(str(record.seq)):
            self.assertIsInstance(gene, TRNAGene)
    
    def test_tmrna(self):
        record = data.load_record("CP001621.fna.gz")
        finder = RNAFinder(translation_table=11, tmrna=True, trna=False)
        for gene in finder.find_rna(str(record.seq)):
            self.assertIsInstance(gene, TMRNAGene)

    def test_ps95(self):
        record = data.load_record("CP001621.fna.gz")
        lines = data.load_text("CP001621.ps95.txt").splitlines()
        
        finder = RNAFinder(translation_table=11, threshold_scale=0.95)
        genes = finder.find_rna(str(record.seq))

        for gene, expected in itertools.zip_longest(genes, batched(lines[2:], 3)):
            self.assertIsNotNone(gene)
            self.assertIsNotNone(expected)
            result, seq, ss = expected
            if gene.type == "tRNA":
                matched = _TRNA_RX.match(result)
                _, aa, complement, begin, end, energy, offset, anticodon = matched.groups()
                self.assertEqual(gene.amino_acid, aa)
                self.assertEqual(gene.begin, int(begin))
                self.assertEqual(gene.end, int(end))
                self.assertEqual(gene.anticodon_offset, int(offset))
                self.assertEqual(gene.anticodon_length, len(anticodon))
                self.assertEqual(gene.anticodon, anticodon)
                self.assertEqual(gene.strand, -1 if complement == "c" else +1)
                self.assertAlmostEqual(gene.energy, float(energy), places=1)
                self.assertEqual(gene.sequence().lower(), seq)
            elif gene.type == "tmRNA":
                matched = _TMRNA_RX.match(result)
                _, complement, begin, end, energy, orf_start, orf_end, peptide = matched.groups()
                self.assertEqual(gene.begin, int(begin))
                self.assertEqual(gene.end, int(end))
                self.assertEqual(gene.orf_offset, int(orf_start))
                self.assertEqual(gene.orf_offset + gene.orf_length, int(orf_end))
                self.assertEqual(gene.peptide(), peptide)
                self.assertEqual(gene.strand, -1 if complement == "c" else +1)
                self.assertAlmostEqual(gene.energy, float(energy), places=1)
                # self.assertEqual(gene.sequence().lower(), seq) # TODO


import typing
from typing import Literal, Union, Generic
from typing_extensions import Buffer

__version__: str

ARAGORN_VERSION: str
TRANSLATION_TABLES: typing.Set[int]

class Gene:
    @property
    def type(self) -> str: ...
    @property
    def begin(self) -> int: ...
    @property
    def end(self) -> int: ...
    @property
    def length(self) -> int: ...
    @property
    def strand(self) -> Literal[1, -1]: ...
    @property
    def energy(self) -> float: ...
    @property
    def raw_energy(self) -> float: ...
    def sequence(self) -> str: ...
 
class TRNAGene(Gene):
    @property
    def type(self) -> Literal["tRNA"]: ...
    @property
    def amino_acid(self) -> str: ...
    @property
    def amino_acids(self) -> Union[Tuple[str], Tuple[str, str]]: ...
    @property
    def anticodon(self) -> str: ...
    @property
    def anticodon_offset(self) -> int: ...
    @property
    def anticodon_length(self) -> Literal[2, 3, 4]: ...

class TMRNAGene(Gene):
    @property
    def type(self) -> Literal["tmRNA"]: ...
    @property
    def permuted(self) -> bool: ...
    @property
    def orf_offset(self) -> int: ...
    @property
    def orf_length(self) -> int: ...
    def orf(self, include_stop: bool = True) -> str: ...
    def peptide(self, include_stop: bool = True) -> str: ...


class Cursor:
    def __init__(self, obj: Union[str, bytes, bytearray, Buffer]) -> None: ...


G = typing.TypeVar("G", bound=Gene)

class RNAFinder(Generic[G]):
    @typing.overload
    def __init__(
        self: RNAFinder[TRNAGene],
        translation_table: int = 1,
        *,
        tmrna: Literal[False],
        trna: Literal[True] = True,
        linear: bool = False,
        threshold_scale: float = 1.0,
    ) -> None: ...
    @typing.overload
    def __init__(
        self: RNAFinder[TMRNAGene],
        translation_table: int = 1,
        *,
        trna: Literal[False],
        tmrna: Literal[True] = True,
        linear: bool = False,
        threshold_scale: float = 1.0,
    ) -> None: ...
    @typing.overload
    def __init__(
        self: RNAFinder[Union[TRNAGene, TMRNAGene]],
        translation_table: int = 1,
        *,
        trna: bool = True,
        tmrna: bool = True,
        linear: bool = False,
        threshold_scale: float = 1.0,
    ) -> None: ...
    def __repr__(self) -> str: ...
    @property
    def translation_table(self) -> int: ...
    @property
    def trna(self) -> bool: ...
    @property
    def tmrna(self) -> bool: ...
    @property
    def linear(self) -> bool: ...
    @property
    def threshold_scale(self) -> float: ...
    @threshold_scale.setter
    def threshold_scale(self, threshold_scale: float) -> None: ...
    def find_rna(
        self,
        sequence: Union[str, bytes, bytearray, Buffer],
    ) -> List[G]: ...



#!/usr/bin/env python3

import csv
import os
import subprocess as sp
import typing
from argparse import ArgumentParser
from collections import namedtuple
from pathlib import Path

from Bio import SeqIO
from xopen import xopen

import pyaragorn


class Gene(typing.NamedTuple):
    id: str
    kind: str
    start: int
    stop: int
    strand: int


def aragorn_predict(genome: Path, txt_file: Path, aragorn_bin: Path, linear: bool, transl_table: int = 11, ps: float = 100.0):
    """
    Predict CDS with aragorn
    :param genome: Path to the genome (fasta)
    :param train_file: Path to the training file
    :param txt_file: Path to text output file
    :param aragorn_bin: Path to the aragorn binary
    :param closed: Closed ends.
    :param transl_table: Translation table
    :param ps: Scoring threshold percentage (default 100.0)
    """
    cmd = [
        aragorn_bin,
        '-o', str(txt_file),
        '-wa',
        '-e',
        '-l' if linear else '-c',
        f'-ps{ps}',
        f'-gc{transl_table}',
        str(genome),
    ]

    proc = sp.run(
        cmd,
        env=os.environ.copy(),
        stdout=sp.PIPE,
        stderr=sp.PIPE,
        universal_newlines=True
    )
    if proc.returncode != 0:
        print('stdout=\'%s\', stderr=\'%s\'', proc.stdout, proc.stderr)
        raise Exception(f'aragorn error! error code: {proc.returncode}')


def parse_txt_output(txt_path: Path, pyaragorn_processed: bool = False) -> set[Gene]:
    """
    Parse the aragorn output.
    :param gff_path: Path to the GFF output file
    :param pyaragorn_processed: pyaragorn output
    :return: Set of namedtuples of CDS coordinates
    """
    cdss: set[Gene] = set()
    contig_id = None
    with txt_path.open() as fh:
        for line in fh:
            if line[0] == '>':
                contig_id = line[1:].split()[0]
                fh.readline()
            else:
                #print(line)
                (n, kind, coords, energy, x, y) = line.strip().split()
                if coords.startswith("c"):
                    strand = -1
                    coords = coords[1:]
                else:
                    strand = 1
                start, stop = map(int, coords[1:-1].split(","))
                gene: namedtuple = Gene(contig_id,
                                        kind,
                                        int(start),
                                        int(stop),
                                        strand)
                cdss.add(gene)
    return cdss



def pyaragorn_predict(rna_finder: pyaragorn.RNAFinder, genome: Path) -> set[Gene]:
    """
    Predict all genes using pyaragorn.
    :param orf_finder: Trained pyaragorn OrfFinder instance
    :param genome: Path to the contig (fasta)
    :return:
    """
    cdss: set[Gene] = set()
    with xopen(genome, mode='r') as f:
        for record in SeqIO.parse(f, 'fasta'):
            for prediction in rna_finder.find_rna(bytes(record.seq)):
                kind = prediction.type
                if prediction.type == "tRNA":
                    if prediction.amino_acid == "???":
                        kind = f"{kind}-?({'|'.join(prediction.amino_acids)})"
                    else:
                        kind = f"{kind}-{prediction.amino_acid}"
                elif prediction.type == "tmRNA":
                    mask = [["", "p"], ["*", "p*"]]
                    kind = "tmRNA{}".format(mask[prediction.permuted][prediction.energy < 100.0])
                gene: namedtuple = Gene(record.id,
                                        kind,
                                        prediction.begin,
                                        prediction.end,
                                        prediction.strand)
                cdss.add(gene)
    return cdss


def compare_mismatches(aragorn_genes: set[Gene], pyaragorn_genes: set[Gene]) -> list:
    """
    Compare aragorn/pyaragorn mismatches to find related hits.
    :param aragorn_genes: Mismatched aragorn genes
    :param pyaragorn_genes: Mismatched pyaragorn genes
    :return: Related hits
    """
    ordered_pairs: list = []
    empty_gene = Gene(None, None, None, None, None)

    longer, shorter = (aragorn_genes, pyaragorn_genes) if len(aragorn_genes) >= len(pyaragorn_genes) else (pyaragorn_genes, aragorn_genes)
    aragorn_more_hits = True if len(aragorn_genes) >= len(pyaragorn_genes) else False

    for main_cds in longer:
        for candidate_cds in shorter:
            if main_cds.id == candidate_cds.id and main_cds.strand == candidate_cds.strand:  # matching contig AND strand
                if main_cds.stop == candidate_cds.stop or main_cds.start == candidate_cds.start:  # matching stop or start
                    mismatch = {'aragorn': main_cds,
                                'pyaragorn': candidate_cds,
                                'strand': main_cds.strand} if aragorn_more_hits else {'aragorn': candidate_cds,
                                                                                       'pyaragorn': main_cds,
                                                                                       'strand': main_cds.strand}
                    ordered_pairs.append(mismatch)
                    break
        else:
            mismatch = {'aragorn': main_cds,
                        'pyaragorn': empty_gene,
                        'strand': main_cds.strand} if aragorn_more_hits else {'aragorn': empty_gene,
                                                                               'pyaragorn': main_cds,
                                                                               'strand': main_cds.strand}
            ordered_pairs.append(mismatch)

    return ordered_pairs


def parse_arguments():
    """
    Argument parser
    :return: Command line arguments.
    """
    parser = ArgumentParser(description='Compare CDS predictions of aragorn to the predictions of pyaragorn and save the differences in a TSV file.')
    parser.add_argument('--genome', '-g', type=Path, nargs='+', action='append', help='Input genomes (/some/path/*.fasta)')
    parser.add_argument('--aragorn', '-a', type=Path, default=Path('aragorn'),
                        help='Path to a newly compiled aragorn binary.')
    parser.add_argument('--linear', '-l', action='store_true', help='Closed ends. Do not allow genes to run off edges.')
    parser.add_argument('--output', '-o', type=Path, default=Path('./'), help='Output path (default="./comparison")')
    parser.add_argument('--ps', '-p', type=float, default=100.0, help='Scoring threshold percentage (default=100.0)')
    args = parser.parse_args()
    return args.genome, args.aragorn, args.linear, args.output, args.ps


def main():
    genomes, aragorn_bin, linear, out_path, ps = parse_arguments()
    out_path.resolve()
    print(f'Genomes linear={linear} ps={ps}')

    tsv_file = open(out_path.joinpath('mismatches.tsv'), mode='w')
    tsv_writer = csv.writer(tsv_file, delimiter='\t', lineterminator='\n')
    header: list[str] = ['genome', 'strand', 'aragorn_kind', 'aragorn_start', 'aragorn_stop', 'pyaragorn_kind', 'pyaragorn_start', 'pyaragorn_stop']
    tsv_writer.writerow(header)

    for genome in genomes[0]:
        genome.resolve()
        prefix: str = str(genome).split('/')[-1].split('.')[0]

        tmp_path = out_path.joinpath(f'tmp/{prefix}')
        try:
            os.makedirs(tmp_path)
        except FileExistsError:
            pass

        # Predict aragorn
        aragorn_txt_file = tmp_path.joinpath(f'{prefix}.aragorn.txt')
        aragorn_predict(genome=genome,
                         txt_file=aragorn_txt_file,
                         aragorn_bin=aragorn_bin,
                         linear=linear,
                         ps=ps)
        aragorn_genes: set[dict] = parse_txt_output(txt_path=aragorn_txt_file)
        # Predict pyaragorn
        rna_finder = pyaragorn.RNAFinder(11, linear=linear, threshold_scale=ps/100.0)
        pyaragorn_genes: set[dict] = pyaragorn_predict(rna_finder=rna_finder,
                                                       genome=genome)

        print(f'Hits genome={prefix}: aragorn={len(aragorn_genes)}, pyaragorn={len(pyaragorn_genes)}, equal={aragorn_genes == pyaragorn_genes}')

        if aragorn_genes != pyaragorn_genes:
            difference_aragorn = aragorn_genes.difference(pyaragorn_genes)
            difference_pyaragorn = pyaragorn_genes.difference(aragorn_genes)
            ordered_pairs = compare_mismatches(difference_aragorn, difference_pyaragorn)

            for mismatch in ordered_pairs:
                # genome, strand, aragorn_start, aragorn_stop, aragorn_edge, pyaragorn_start, pyaragorn_stop, pyaragorn_edge
                line = [prefix,
                        mismatch['strand'],
                        mismatch['aragorn'].kind,
                        mismatch['aragorn'].start,
                        mismatch['aragorn'].stop,
                        mismatch['pyaragorn'].kind,
                        mismatch['pyaragorn'].start,
                        mismatch['pyaragorn'].stop]

                tsv_writer.writerow(line)

    tsv_file.close()


if __name__ == '__main__':
    main()



from libc.stdio cimport FILE

cdef extern from "aragorn.h" nogil:

    const size_t STRLEN
    const size_t STRLENM1
    const size_t SHORTSTRLEN
    const size_t SHORTSTRLENM1

    const size_t NS

    const size_t MAXGCMOD
    const size_t MAMMAL_MT
    const size_t NGENECODE

    const size_t MININTRONLEN
    const size_t MAXINTRONLEN
    const size_t MINCTRNALEN
    const size_t MAXCTRNALEN
    const size_t MINTRNALEN
    const size_t MAXTRNALEN
    const size_t MAXETRNALEN

    const size_t MINTAGDIST
    const size_t MAXTAGDIST

    const size_t MAXTMRNALEN
    const size_t TSWEEP
    const size_t WRAP
    const size_t NPTAG
    const size_t MAXAGENELEN

    const size_t NA
    const size_t ND
    const size_t NT
    const size_t NH
    const size_t NTH
    const size_t NC
    const size_t NGFT
    const size_t NTAG
    const size_t NTAGMAX
    const size_t LSEQ
    const size_t ATBOND

    cdef bint space(char c)
    cdef long int sq(long int pos)

    cdef enum base:
        INSERT
        TERM
        Adenine
        Cytosine
        Guanine
        Thymine
        AMBIG
        NOBASE

    cdef enum data_type:
        FASTA
        GENBANK

    cdef enum gencode:
        METAZOAN_MT
        STANDARD
        VERTEBRATE_MT

    cdef enum gene_type:
        noGENE
        tRNA
        tmRNA
        srpRNA
        rRNA
        CDS

    ctypedef struct annotated_gene:
        long int start
        long int stop
        int comp
        long int antistart
        long int antistop
        gene_type genetype
        bint pseudogene
        bint permuted
        bint detected
        char species[SHORTSTRLEN]

    ctypedef struct data_set:
        char filename[80]
        FILE *f
        char seqname[STRLEN]
        bint bugmode
        data_type datatype
        double gc
        long int filepointer
        long int ps
        long int psmax
        long int seqstart
        long int seqstartoff
        long int nextseq
        long int nextseqoff
        int ns
        int nf
        long int aseqlen
        int nagene[NS]
        annotated_gene gene[NGFT]

    ctypedef struct gene:
        char name[100]
        int seq[MAXTRNALEN + 1]
        int eseq[MAXETRNALEN + 1]
        int *ps
        int nbase
        int comp
        long int start
        long int stop
        int astem1
        int astem2
        int aatail
        int spacer1
        int spacer2
        int dstem
        int dloop
        int cstem
        int cloop
        int intron
        int nintron
        int anticodon
        int var
        int varbp
        int tstem
        int tloop
        int genetype
        double energy
        int asst
        int tps
        int tpe
        int annotation
        int annosc

    ctypedef struct csw:
        char genetypename[NS][10]
        FILE *f
        int batch
        int batchfullspecies
        int repeatsn
        int trna
        int tmrna
        int srprna
        int cds
        int mtrna
        int tvloop
        int cloop7
        int peptide
        int geneticcode
        int ngcmod
        int gcmod[MAXGCMOD]
        int gcfix
        int discrim
        int extastem
        int tarm
        int tagthresh
        int tarmlength
        int showconfig
        int libflag
        bint verbose
        int linear
        int both
        int reportpseudogenes
        int energydisp
        int secstructdisp
        int seqdisp
        int aataildisp
        int aataildiv
        int sp1max
        int sp2min
        int sp2max
        int mtxdetect
        int mtcdsscan
        int mtcompov
        int matchacceptor
        int maxintronlen
        int minintronlen
        int minintronlenreport
        int ioverlay
        int ifixedpos
        int ireportminintronlen
        int tmstrict
        int iamismatch
        int loffset
        int roffset
        long int start
        int comp
        gene* genes
        int genespace
        int srpspace
        int ngene[NS]
        int nps
        int annotated
        int dispmatch
        int updatetmrnatags
        int tagend
        int trnalenmisthresh
        int tmrnalenmisthresh
        int nagene[NS]
        int nafn[NS]
        int nafp[NS]
        int natfpd
        int natfptv
        int lacds
        int ldcds
        long int nabase
        double reportpsthresh
        double threshlevel
        double trnathresh
        double ttscanthresh
        double ttarmthresh
        double tdarmthresh
        double tastemthresh
        double tascanthresh
        double mttthresh
        double mtdthresh
        double mtdtthresh
        double mttarmthresh
        double mtdarmthresh
        double tmrnathresh
        double tmathresh
        double tmcthresh
        double tmcathresh
        double tmrthresh
        double srpthresh
        double cdsthresh
        double eref[NS]
        int tmrna_struct[200]

    int[3][6][6] lbp
    int[6][6] bp
    int[6][6] wbp
    int[6][6] wcbp
    int[6][6] gc
    int[6][6] gt
    int[6][6] at
    int[6][6] tt
    int[6][6] stemterm
    int[6][6] aastemterm
    int[6][6] ggstemterm
    int[6][6] assymst
    int[6][6] assymat
    int[6][6] stackbp
    int[6][6] ggstackbp
    int[6][6] ggbp
    int[6][6] gabp
    int[6][6] assymagbp
    int[6][6] stembp
    int[6][6] ggstembp
    int[6][6] gastembp
    int[6][6] vbp
    int[256] map
    # int[mtNTM][4] tandemid
    # double[mtNTM] tandem_em
    # double[6][6] send_em
    # double[6][6] ssend_em
    # int[6][6] neighbour_map
    # double[2][6][6] neighbour_em
    # unsigned int[6][6] btmap
    # double[6][6] bem
    # int[3][64][6] mt_discrim
    # char[NAMINOACID + 1] aapolarity
    # char[NAMINOACID + 1] aaletter
    # char[NAMINOACID][20] aaname
    # char[4] ambig_aaname
    # int[NGENECODE][64] aamap
    # char[NHELPLINE][81] helpmenu
    # tmrna_tag_entry[NTAGMAX] tagdatabase

    char* aa(int* anticodon, csw* sw)
    void change_thresholds(csw *sw, double psthresh)
    void bopt_fastafile(data_set *d, csw *sw)
    void batch_gene_set(data_set* d, int nt, csw* sw)
    char cbase(int c)
    char cpbase(int c)
    int gene_sort(data_set *d, int nt, int* sort, csw *sw)
    void init_gene(gene* ts, int nstart, int nstop)
    char ltranslate(int *codon, gene *t, csw *sw)
    int move_forward(data_set *d)
    double nenergy(gene *t, csw *sw)
    void remove_overlapping_trna(data_set *d, int nt, csw *sw)
    void report_new_tmrna_tags(csw *sw)
    void sense_switch(int *seq1, int *seq2, int lseq)
    int seq_init(data_set *d, csw *sw)
    int seqlen(gene* t)
    int tmioptimise(data_set *d, int *seq, int lseq, int nts, csw *sw)
    void update_tmrna_tag_database(gene* ts, int nt, csw *sw)


import argparse
import copy
import os
import sys
import itertools
import pathlib
import tempfile
import subprocess

import pycparser
import pycparser.c_ast
from pycparser.c_ast import NodeVisitor
from pycparser.c_generator import CGenerator

parser = argparse.ArgumentParser()
parser.add_argument("-i", "--input", type=pathlib.Path, required=True, help="path to the source file")
parser.add_argument("-o", "--output", type=pathlib.Path, required=True, help="path to the output folder")
parser.add_argument("--cpp", default="cpp", help="name of the C preprocessor executable")
args = parser.parse_args()

# record contents, includes & defines
contents = ["typedef void FILE;\n"]  # needed for pycparser to be happy about 
includes = ["#include <stdint.h>\n"]
defines = []

# load file, extract defines and includes
with args.input.open() as f:
    for line in map(str.lstrip, f):
        if line.startswith(("#define", "#ifndef", "#endif")):
            defines.append(line)
        elif line.startswith("#include"):
            includes.append(line)
        else:
            contents.append(line)

# preprocess contents and load AST
with tempfile.TemporaryDirectory() as folder:
    source = pathlib.Path(folder).joinpath("aragorn.c")
    with source.open("w") as tmp:
        tmp.writelines(contents)
    flag = "/EP" if os.name == "nt" else "-E"
    proc = subprocess.run([args.cpp, flag, os.fspath(source)], stdout=subprocess.PIPE)
    proc.check_returncode()
    parser = pycparser.CParser()
    ast = parser.parse("\n".join(proc.stdout.decode().splitlines()))

# record typedefs, constant arrays, functions
typedefs = []
functions = []
arrays = []
data = []

# process AST, extract functions, constant arrays, typedefs
for i, (_, node) in enumerate(itertools.islice(ast.children(), 1, None)):
    # extract typedef
    if isinstance(node, pycparser.c_ast.Typedef):
        typedefs.append(node)
    # extract constant array
    elif isinstance(node, pycparser.c_ast.Decl) and isinstance(node.type, pycparser.c_ast.ArrayDecl):
        arrays.append(node)
    # extract function definition
    elif isinstance(node, pycparser.c_ast.FuncDef):
        functions.append(node)

# --- Transform AST ------------------------------------------------------------

# add "genes" field to the `csw` struct
node = next(ty for ty in typedefs if ty.name == "csw")
node.type.type.decls.insert(1, 
    pycparser.c_ast.Decl(
        name='genes', 
        quals=[], 
        align=[], 
        storage=[], 
        funcspec=[], 
        init=None,
        bitsize=None, 
        type=pycparser.c_ast.TypeDecl(
            declname='genes',
            quals=[],
            align=None,
            type=pycparser.c_ast.IdentifierType(names=['gene*'])
        )
    )
)

# replace all uses of ts
class FunctionPatcher(NodeVisitor):

    @property
    def _replacement(self):
        return pycparser.c_ast.StructRef(
            name=pycparser.c_ast.ID(name='sw'),
            type='->',
            field=pycparser.c_ast.ID(name='genes')
        )

    def visit_Assignment(self, node):
        if isinstance(node.lvalue, pycparser.c_ast.ID) and node.lvalue.name == "ts":
            node.lvalue = self._replacement
        super().generic_visit(node)

    def visit_ArrayRef(self, node):
        if isinstance(node.name, pycparser.c_ast.ID) and node.name.name == "ts":
            node.name = self._replacement
        super().generic_visit(node)

    def visit_FuncCall(self, node):
        if node.name.name == "init_gene":
            node.args.exprs.insert(0, self._replacement)
        elif node.name.name == "nearest_tmrna_gene":
            node.args.exprs.append(pycparser.c_ast.ID(name="sw"))
        elif node.args is not None:
            for i, expr in enumerate(node.args.exprs):
                if isinstance(expr, pycparser.c_ast.ID) and expr.name == "ts":
                    node.args.exprs[i] = self._replacement

        super().generic_visit(node)

    def visit_BinaryOp(self, node):
        if isinstance(node.left, pycparser.c_ast.ID) and node.left.name == "ts":
            node.left = pycparser.c_ast.StructRef(
                name=pycparser.c_ast.ID(name='sw'),
                type='->',
                field=pycparser.c_ast.ID(name='genes')
            )
        super().generic_visit(node)

    def visit_Cast(self, node):
        if isinstance(node.expr, pycparser.c_ast.ID) and node.expr.name == "ts":
            node.expr = self._replacement
        super().generic_visit(node)

for node in functions:
    if node.decl.name != "init_gene":
        FunctionPatcher().visit(node)

# add sw argument to `nearest_tmrna_gene`
node = next(f for f in functions if f.decl.name == "nearest_tmrna_gene")
node.decl.type.args.params.append(
    pycparser.c_ast.Decl(
        name='sw', 
        quals=[], 
        align=[], 
        storage=[], 
        funcspec=[], 
        init=None,
        bitsize=None, 
        type=pycparser.c_ast.TypeDecl(
            declname='sw',
            quals=[],
            align=None,
            type=pycparser.c_ast.IdentifierType(names=['csw*'])
        )
    )
)

# add "ts" argument to `init_gene`
node = next(f for f in functions if f.decl.name == "init_gene")
node.decl.type.args.params.insert(0,
    pycparser.c_ast.Decl(
        name='ts', 
        quals=[], 
        align=[], 
        storage=[], 
        funcspec=[], 
        init=None,
        bitsize=None, 
        type=pycparser.c_ast.TypeDecl(
            declname='ts',
            quals=[],
            align=None,
            type=pycparser.c_ast.IdentifierType(names=['gene*'])
        )
    )
)

# reduce size of aamap array (int to int8_t)
aamap = next(a for a in arrays if a.name == "aamap")
aamap.type.type.type.type = pycparser.c_ast.ID(name="int8_t")

# extract the map array from move_forward
move_forward = next(f for f in functions if f.decl.name == "move_forward")
map_array = move_forward.body.block_items.pop(2)
map_array.storage = []
arrays.append(map_array)

# remove main function
main = next(f for f in functions if f.decl.name == "main")
functions.remove(main)

# remove helpmenu array and function
helpmenu = next(a for a in arrays if a.name == "helpmenu")
arrays.remove(helpmenu)
aragorn_help_menu = next(f for f in functions if f.decl.name == "aragorn_help_menu")
functions.remove(aragorn_help_menu)


# --- Emit code ----------------------------------------------------------------

args.output.mkdir(exist_ok=True)
gen = CGenerator()

# write common header
with args.output.joinpath("aragorn.h").open("w") as dst:
    dst.write("#ifndef ARAGORN_H\n")
    dst.write("#define ARAGORN_H\n")
    # write includes and defines from the original file
    dst.writelines(includes)
    dst.writelines(defines)
    # write typedefs
    for node in typedefs:
        dst.write(gen.visit(node) + ";\n")
    # write arrays
    for node in arrays:
        node = copy.deepcopy(node)
        node.init = None
        node.storage = ["extern"]
        dst.write(gen.visit(node) + ";\n")
    # write function signatures
    for node in functions:
        node = copy.deepcopy(node.decl)
        node.storage = ["extern"]
        dst.write(gen.visit(node) + ";\n")
    dst.write("#endif\n")


# write common source
with args.output.joinpath("aragorn.c").open("w") as dst:
    dst.write(f'#include "aragorn.h"\n')
    # write function definitions
    for node in functions:
        dst.write(gen.visit(node)+"\n")

# with args.output.joinpath("main.c").open("w") as dst:
#     dst.write('#include "aragorn.h"\n')
#     dst.write(gen.visit(main)+"\n")

# write data file
with args.output.joinpath("data.c").open("w") as dst:
    dst.write('#include "aragorn.h"\n')
    # write arrays
    for node in arrays:
        dst.write(gen.visit(node) + ";\n")



import bz2
import collections
import contextlib
import io
import os



_BZ2_MAGIC = b"BZh"
_GZIP_MAGIC = b"\x1f\x8b"
_XZ_MAGIC = b"\xfd7zXZ"
_LZ4_MAGIC = b"\x04\x22\x4d\x18"
_ZSTD_MAGIC = b"\x28\xb5\x2f\xfd"


@contextlib.contextmanager
def zopen(file, mode='r', encoding=None, errors=None, newline=None):
    with contextlib.ExitStack() as ctx:

        try:
            path = os.fsencode(file)
            file = ctx.enter_context(open(path, mode="rb"))
        except TypeError:
            raise

        peek = file.peek()
        if peek.startswith(_GZIP_MAGIC):
            try:
                from isal import igzip as gzip
            except ImportError:
                import gzip  # type: ignore
            file = ctx.enter_context(gzip.open(file, mode="rb"))
        elif peek.startswith(_BZ2_MAGIC):
            import bz2
            file = ctx.enter_context(bz2.open(file, mode="rb"))
        elif peek.startswith(_XZ_MAGIC):
            import lzma
            file = ctx.enter_context(lzma.open(file, mode="rb"))
        elif peek.startswith(_LZ4_MAGIC):
            try:
                import lz4.frame
            except ImportError as err:
                raise RuntimeError("File compression is LZ4 but lz4 is not installed") from err
            file = ctx.enter_context(lz4.frame.open(file))
        elif peek.startswith(_ZSTD_MAGIC):
            try:
                import zstandard
            except ImportError as err:
                raise RuntimeError("File compression is ZSTD but zstandard is not installed") from err
            decompressor = zstandard.ZstdDecompressor()
            file = decompressor.stream_reader(file)
        if mode == "r":
            file = io.TextIOWrapper(file, encoding=encoding, errors=errors, newline=newline)
        yield file


class Record(collections.namedtuple("Record", ["id", "seq", "description"])):
    pass


def parse(path):
    with contextlib.ExitStack() as ctx:
        try:
            path = os.fsencode(path)
            file = ctx.enter_context(zopen(path, "r"))
        except TypeError:
            file = path

        # parse file
        id_ = None
        seq = []
        for line in file:
            l = line.strip()
            if line.startswith(">"):
                if id_ is not None:
                    yield Record(id_, "".join(seq), desc)
                fields = line[1:].split(maxsplit=1)
                id_ = fields[0] if len(fields) > 0 else ""
                desc = fields[1] if len(fields) > 1 else ""
                seq = []
            elif l:
                seq.append(l)
        if id_ is not None:
            yield Record(id_, "".join(seq), desc)
        elif seq:
            raise ValueError("not in FASTA format")



import contextlib
import gzip
import io

try:
    try:
        from importlib.resources import files
    except ImportError:
        from importlib_resources import files  # type: ignore
except ImportError:
    files = None  # type: ignore

from ..fasta import parse, zopen

@contextlib.contextmanager
def load(name, mode="rb"):
     with zopen(files(__name__).joinpath(name), mode=mode) as src:
        yield src

def load_record(name):
    with load(name, mode="r") as f:
        return next(parse(f))

def load_records(name):
    with load(name, mode="r") as f:
        return list(parse(f))

def load_text(name):
    with load(name, mode="r") as f:
        return f.read()


from . import (
    test_rna_finder,
    test_doctest,
)


def load_tests(loader, suite, pattern):
    suite.addTests(loader.loadTestsFromModule(test_rna_finder))
    test_doctest.load_tests(loader, suite, pattern)
    return suite



# coding: utf-8
# cython: language_level=3, linetrace=True, binding=True

"""Bindings to ARAGORN, a (t|mt|tm)RNA gene finder.

Attributes:
    ARAGORN_VERSION (`str`): The version of ARAGORN currently wrapped
        in PyARAGORN.
    TRANSLATION_TABLES (`set` of `int`): A set containing all the
        translation tables supported by PyARAGORN.

Example:
    PyARAGORN can work on any DNA sequence stored in either a text or a
    byte array. To load a sequence from one of the common sequence formats,
    you can use an external dedicated library such as
    `Biopython <https://github.com/biopython/biopython>`_::

        >>> import gzip
        >>> import Bio.SeqIO
        >>> with gzip.open("CP001621.fna.gz", "rt") as f:
        ...     record = Bio.SeqIO.read(f, "fasta")

    Then use PyARAGORN to find the tRNA genes using the
    bacterial genetic code (translation table 11):

        >>> import pyaragorn
        >>> rna_finder = pyaragorn.RNAFinder(11, trna=True, tmrna=False)
        >>> for gene in rna_finder.find_rna(record.seq.encode()):
        ...     print(gene.anticodon, gene.amino_acid, gene.begin, gene.end)
        tag Leu 87124 87207
        ttt Lys 87210 87285
        ...

    The gene coordinates are 1-indexed, inclusive, similarly to
    `Pyrodigal <https://pyrodigal.readthedocs.io>`_ genes.

References:
    - Laslett, Dean, and Björn Canback.
      “ARAGORN, a program to detect tRNA genes and tmRNA genes in nucleotide
      sequences.” Nucleic acids research vol. 32,1 11-6. 2 Jan. 2004,
      :doi:`10.1093/nar/gkh152`. :pmid:`14704338`. :pmcid:`PMC373265`.
    - Laslett, Dean, and Björn Canbäck.
      “ARWEN: a program to detect tRNA genes in metazoan mitochondrial
      nucleotide sequences.” Bioinformatics (Oxford, England) vol. 24,2
      (2008): 172-5. :doi:`10.1093/bioinformatics/btm573`. :pmid:`18033792`.

"""

from cython.operator cimport postincrement, dereference
from cpython.bytes cimport PyBytes_FromStringAndSize
from cpython.exc cimport PyErr_CheckSignals
from cpython.unicode cimport PyUnicode_AsASCIIString

from libc.stdio cimport FILE, fopen, fdopen, fclose, fprintf, fputc, stdout, stderr
from libc.stdlib cimport calloc, free
from libc.string cimport memcpy
from libc.stdint cimport intptr_t

cimport aragorn
from aragorn cimport csw, data_set, gene

# --- Helpers ------------------------------------------------------------------

cdef extern from * nogil:
    Py_UCS4 PyUnicode_READ(int kind, const void* data, size_t pos)

cdef extern from * nogil:
    """
    void default_sw(csw* sw) {
        csw x = {
            {"tRNA", "tmRNA", "", "", "CDS", "overall"},
            NULL, NULL, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, STANDARD, 0,
            {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
            0, METAZOAN_MT, 1, 0, 5, 5, 1, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0,
            3, 0, 2, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0,
            {0, 0, 0, 0, 0, 0}, 0, 0, 0, 0, NTAG, 10, 30,
            {0, 0, 0, 0, 0, 0}, {0, 0, 0, 0, 0, 0}, {0, 0, 0, 0, 0, 0},
            0, 0, 0, 0, 0L, 100.0, 1.0, tRNAthresh, 4.0, 29.0, 26.0, 7.5, 8.0,
            mtRNAtthresh, mtRNAdthresh, mtRNAdtthresh, -7.9, -6.0, tmRNAthresh,
            14.0, 10.0, 25.0, 9.0, srpRNAthresh, CDSthresh,
            {tRNAthresh, tmRNAthresh, srpRNAthresh, 0.0, CDSthresh},
            {
                45, 45, 45, 45, 45, 45, 45, 45, 45, 45, 45, 45, 45, 45, 45, 45,
                45, 45, 45, 45, 45, 45, 45, 45, 45, 45, 45, 45, 45, 45, 10, 65,
                82, 65, 71, 79, 82, 78, 32, 118, 49, 46, 50, 46, 52, 49, 32,
                32, 32, 68, 101, 97, 110, 32, 76, 97, 115, 108, 101, 116, 116,
                10, 45, 45, 45, 45, 45, 45, 45, 45, 45, 45, 45, 45, 45, 45, 45,
                45, 45, 45, 45, 45, 45, 45, 45, 45, 45, 45, 45, 45, 45, 45, 10,
                TERM
            }
        };
        memcpy(sw, &x, sizeof(csw));
    }
    """
    void default_sw(csw* sw)

cdef inline long int sq(data_set* d, long int pos) nogil:
    return (pos + d.psmax - 1) % d.psmax + 1


# --- Constants ----------------------------------------------------------------

import functools

cdef set _TRANSLATION_TABLES  = set(range(1, 7)) | set(range(9, 17)) | set(range(21, 27)) | {29, 30} | {32, 33}

__version__ = PROJECT_VERSION

TRANSLATION_TABLES = _TRANSLATION_TABLES
ARAGORN_VERSION = PROJECT_ARAGORN_VERSION


# --- Classes ------------------------------------------------------------------

cdef class Gene:
    """A gene identified by ARAGORN.
    """

    cdef gene _gene
    cdef int  _genetic_code

    @staticmethod
    cdef Gene _new_gene(gene* _gene, int _genetic_code):
        cdef Gene obj

        if _gene.genetype == aragorn.tRNA:
            obj = TRNAGene.__new__(TRNAGene)
        elif _gene.genetype == aragorn.tmRNA:
            obj = TMRNAGene.__new__(TMRNAGene)
        else:
            raise NotImplementedError

        memcpy(&obj._gene, _gene, sizeof(gene))
        obj._genetic_code = _genetic_code
        return obj

    def __sizeof__(self):
        return sizeof(self)

    @property
    def type(self):
        return ["tRNA", "tmRNA", "", "", "CDS"][<int> self._gene.genetype]

    @property
    def begin(self):
        """`int`: The sequence coordinate at which the gene begins.

        Hint:
            This coordinate is 1-based, inclusive. To use it to index
            a Python array or string, subtract one.

        """
        return self._gene.start

    @property
    def end(self):
        """`int`: The sequence coordinate at which the gene end.

        Hint:
            This coordinate is 1-based, inclusive. To use it to index
            a Python array or string, subtract one.

        """
        return self._gene.stop

    @property
    def length(self):
        """`int`: The length of the RNA gene.
        """
        return aragorn.seqlen(&self._gene)

    @property
    def strand(self):
        """`int`: *-1* if the gene is on the reverse strand, *+1* otherwise.
        """
        return -1 if self._gene.comp else +1

    @property
    def energy(self):
        """`float`: The approximated normalised energy of the RNA structure.
        """
        cdef csw sw
        default_sw(&sw) # FIXME: should use the same parameters as the
                        #        RNAFinder that produced the gene
        return aragorn.nenergy(&self._gene, &sw)

    @property
    def raw_energy(self):
        """`float`: The un-normalized energy value of the RNA structure."""
        return <double> self._gene.energy

    def sequence(self):
        """Retrieve the full sequence of the RNA gene.
        """
        cdef int       i
        cdef int       l = aragorn.seqlen(&self._gene)
        cdef bytearray b = bytearray(l)
        for i in range(l):
            b[i] = aragorn.cpbase(self._gene.seq[i])
        return b.decode('ascii')


cdef class TRNAGene(Gene):
    """A transfer RNA (tRNA) gene.
    """

    def __repr__(self):
        return (
            f"<TRNAGene begin={self.begin} end={self.end} "
            f"strand={self.strand:+} "
            f"length={self.length} anticodon={self.anticodon!r} "
            f"energy={self.energy:.2f}>"
        )

    @property
    def amino_acid(self):
        """`str`: The 3-letter amino-acid(s) for this gene.

        Hint:
            If the anticodon loop contains 6 or 8 bases, ``???`` is
            returned.

        """
        cdef csw sw
        cdef int* s = self._gene.seq + self._gene.anticodon
        (<int*> &sw.geneticcode)[0] = self._genetic_code
        if self._gene.cloop == 6 or self._gene.cloop == 8:
            return "???"
        else:
            return aragorn.aa(s, &sw).decode('ascii')

    @property
    def amino_acids(self):
        """`tuple` of `str`: All possible 3-letter amino-acids for this gene.

        Hint:
            If the anticodon loop contains 6 or 8 bases, a tuple of two
            amino-acid is returned, otherwise a tuple with a single element
            is returned.

        """
        cdef csw sw
        cdef int* s = self._gene.seq + self._gene.anticodon
        (<int*> &sw.geneticcode)[0] = self._genetic_code
        if self._gene.cloop == 6:
            return (
                aragorn.aa(s - 1, &sw).decode('ascii'),
                aragorn.aa(s, &sw).decode('ascii'),
            )
        elif self._gene.cloop == 8:
            return (
                aragorn.aa(s, &sw).decode('ascii'),
                aragorn.aa(s + 1, &sw).decode('ascii')
            )
        else:
            return aragorn.aa(s, &sw).decode('ascii')

    @property
    def anticodon(self):
        """`str`: The anticodon of the tRNA gene.
        """
        cdef tuple c
        cdef int*  s = self._gene.seq + self._gene.anticodon
        if self._gene.cloop == 6:
            c = ( aragorn.cbase(s[0]), aragorn.cbase(s[1]) )
        elif self._gene.cloop == 8:
            c = ( aragorn.cbase(s[0]), aragorn.cbase(s[1]), aragorn.cbase(s[2]), aragorn.cbase(s[3]) )
        else:
            c = ( aragorn.cbase(s[0]), aragorn.cbase(s[1]), aragorn.cbase(s[2]) )
        return ''.join(map(chr, c))

    @property
    def anticodon_offset(self):
        """`int`: The offset in the gene at which the anticodon starts.
        """
        cdef int x = 1 + self._gene.anticodon
        if self._gene.nintron > 0 and self._gene.intron <= self._gene.anticodon:
            x += self._gene.nintron
        return x

    @property
    def anticodon_length(self):
        """`int`: The length of the anticodon (in nucleotides).
        """
        if self._gene.cloop == 6:
            return 2
        elif self._gene.cloop == 8:
            return 4
        else:
            return 3


cdef class TMRNAGene(Gene):
    """A transfer-messenger RNA (tmRNA) gene.

    Example:
        >>> rna_finder = pyaragorn.RNAFinder(11, trna=False, tmrna=True)
        >>> tmrna = rna_finder.find_rna(str(record.seq))[0]
        >>> tmrna.begin, tmrna.end
        (198037, 198447)
        >>> tmrna.peptide()
        'AEKNEENFEMPAFMINNASAGANYMFA**'

    """

    def __repr__(self):
        return (
            f"<TMRNAGene begin={self.begin} end={self.end} "
            f"strand={self.strand:+} "
            f"length={self.length} orf_length={self.orf_length} "
            f"energy={self.energy:.2f}>"
        )

    @property
    def permuted(self):
        """`bool`: Whether this tmRNA gene is a permuted gene.
        """
        return self._gene.asst != 0

    @property
    def orf_offset(self):
        """`int`: The offset in the gene at which the open-reading frame starts.
        """
        return self._gene.tps + 1

    @property
    def orf_length(self):
        """`int`: The length of the open-reading frame (in nucleotides).
        """
        cdef int  tpe    = self._gene.tpe
        cdef int* se     = (self._gene.eseq + tpe) + 1
        cdef int* sb     = (self._gene.eseq + self._gene.tps)
        cdef int  stride = 3

        cdef csw sw
        (<int*> &sw.geneticcode)[0] = self._genetic_code

        while aragorn.ltranslate(se, &self._gene, &sw) == ord('*'):
            se += stride
            tpe += stride

        return tpe - self._gene.tps

    def orf(self, include_stop=True):
        """Retrieve the open-reading frame of the mRNA-like region.

        Arguments:
            include_stop (`bool`): Whether or not to include the STOP codons
                in the returned nucleotide sequence. Defaults to `True`.

        Returns:
            `str`: The sequence of the mRNA-like region in the tmRNA
            gene, optionally without STOP codons.

        """
        cdef int  tpe    = self._gene.tpe
        cdef int* se     = (self._gene.eseq + tpe) + 1
        cdef int* sb     = (self._gene.eseq + self._gene.tps)
        cdef int  stride = 3 if include_stop else -3

        cdef csw sw
        (<int*> &sw.geneticcode)[0] = self._genetic_code

        while aragorn.ltranslate(se, &self._gene, &sw) == ord('*'):
            se += stride
            tpe += stride

        cds = bytearray()
        while sb < se:
            cds.append(aragorn.cpbase(sb[0]))
            sb += 1

        return cds.decode('ascii')

    def peptide(self, include_stop=True):
        """Retrieve the peptide sequence of the mRNA-like region.

        Arguments:
            include_stop (`bool`): Whether or not to include the STOP codons
                in the returned peptide sequence. Defaults to `True`.

        Returns:
            `str`: The translation of the mRNA-like region of the tmRNA
            gene, optionally without STOP codons.

        """
        cdef int  tpe    = self._gene.tpe
        cdef int* se     = (self._gene.eseq + tpe) + 1
        cdef int* sb     = (self._gene.eseq + self._gene.tps)
        cdef int  stride = 3 if include_stop else -3

        cdef csw sw
        (<int*> &sw.geneticcode)[0] = self._genetic_code

        while aragorn.ltranslate(se, &self._gene, &sw) == ord('*'):
            se += stride
            tpe += stride

        peptide = bytearray()
        while sb < se:
            peptide.append(aragorn.ltranslate(sb, &self._gene, &sw))
            sb += 3

        return peptide.decode('ascii')


cdef class Cursor:
    cdef object                 obj
    cdef const unsigned char[:] data
    cdef int                    kind
    cdef size_t                 length
    cdef data_set               ds

    def __init__(self, obj):
        if isinstance(obj, str):
            obj = PyUnicode_AsASCIIString(obj)

        # get a memoryview to the data contents
        self.data = obj
        self.length = self.data.shape[0]

        # keep a reference to the data source
        self.obj = obj

        # reinitialize dataset book-keeping
        self.ds.filepointer = 0
        self.ds.ns = 0
        self.ds.nf = 0
        self.ds.nextseq = 0
        self.ds.nextseqoff = 0
        self.ds.seqstart = 0
        self.ds.seqstartoff = 0
        self.ds.ps = 0
        self.ds.psmax = self.length

        # count GC%
        self.ds.gc = self._gc()

    cdef int _forward(self) noexcept nogil:
        cdef char x
        cdef int  base

        if self.ds.ps >= self.ds.psmax:
            return <int> aragorn.base.TERM

        x = self.data[self.ds.ps]
        if x >= 128:
            return <int> aragorn.base.NOBASE

        base = aragorn.map[x]
        if base >= <int> aragorn.base.Adenine:
            self.ds.ps += 1
            return base
        else:
            return <int> aragorn.base.NOBASE

    cdef double _gc(self) noexcept nogil:
        cdef long i
        cdef char x
        cdef int  base
        cdef long ngc  = 0
        cdef long ps   = 0

        for i in range(self.length):
            x = self.data[i]
            base = aragorn.map[x]
            if base == -1:
                break
            ngc += (base == <int> aragorn.base.Cytosine) or (base == <int> aragorn.base.Guanine)
            ps += 1

        return <double> ngc / <double> ps


cdef class RNAFinder:
    """A configurable RNA gene finder.
    """
    cdef csw _sw

    def __init__(
        self,
        int translation_table = 1,
        *,
        bint trna = True,
        bint tmrna = True,
        bint linear = False,
        double threshold_scale = 1.0,
    ):
        """__init__(self, translation_table=1, *, trna=True, tmrna=True, linear=False, threshold_scale=1.0)\n--\n

        Create a new RNA finder.

        Arguments:
            translation_table (`int`, optional): The translation table to
                use. Check the :wiki:`List of genetic codes` page
                listing all genetic codes for the available values, or
                the :attr:`pyaragorn.TRANSLATION_TABLES` constant for allowed
                values.

        Keyword Arguments:
            trna (`bool`): Enable detection of tRNA genes. Set to `False` to
                disable.
            trmna (`bool`): Enable detection of tmRNA genes. Set to `False` to
                disable.
            linear (`bool`): Set to `True` to assume that the given sequences
                have linear topology (no closed genomes).
            threshold_scale (`float`, optional): Rescale scoring thresholds
                from the default levels. Defaults to 1.0 (no rescaling). Set
                to e.g. 0.95 to report possible pseudogenes by lowering
                the threshold by 5%.

        .. versionadded:: 0.3.0
            The ``threshold_scale`` keyword argument.

        """
        default_sw(&self._sw)
        self._sw.trna = trna
        self._sw.tmrna = tmrna
        self._sw.linear = linear
        self._sw.f = stdout
        self._sw.verbose = False #True
        self.threshold_scale = threshold_scale

        if translation_table not in _TRANSLATION_TABLES:
            raise ValueError(f"invalid translation table: {translation_table!r}")
        self._sw.geneticcode = translation_table

    def __reduce__(self):
        return functools.partial(
            type(self),
            translation_table=self.translation_table,
            trna=self.trna,
            tmrna=self.tmrna,
            linear=self.linear,
            threshold_scale=self.threshold_scale,
        ), ()

    def __repr__(self):
        cdef str  ty   = type(self).__name__
        cdef list args = []

        if self.translation_table != 1:
            args.append(f"{self.translation_table!r}")
        if not self.trna:
            args.append(f"trna={self.trna!r}")
        if not self.tmrna:
            args.append(f"tmrna={self.tmrna!r}")
        if self.linear:
            args.append(f"linear={self.linear!r}")
        if self.threshold_scale != 1.0:
            args.append(f"threshold_scale={self.threshold_scale!r}")
        return f"{ty}({', '.join(args)})"

    @property
    def translation_table(self):
        """`int`: The translation table in use by this object.
        """
        return self._sw.geneticcode

    @property
    def trna(self):
        """`bool`: Whether tRNA detection is enabled.
        """
        return bool(self._sw.trna)

    @property
    def tmrna(self):
        """`bool`: Whether tmRNA detection is enabled.
        """
        return bool(self._sw.tmrna)

    @property
    def linear(self):
        """`bool`: Whether input sequences are assumed to have linear topology.
        """
        return bool(self._sw.linear)

    @property
    def threshold_scale(self):
        """`float`: The scale used to change the default thresholds.

        .. versionadded:: 0.3.0

        """
        return self._sw.threshlevel

    @threshold_scale.setter
    def threshold_scale(self, double threshold_scale):
        if threshold_scale <= 0.0:
            raise ValueError(f"threshold_scale must be positive (got {threshold_scale!r})")
        aragorn.change_thresholds(&self._sw, threshold_scale)

    def find_rna(self, object sequence):
        """Find RNA genes in the input DNA sequence.

        Arguments:
            sequence (`str` or buffer): The nucleotide sequence to process,
                either as a string of nucleotides (upper- or lowercase), or
                as an object implementing the buffer protocol.

        Returns:
            `list` of `~pyaragorn.Gene`: A list of `~pyaragorn.Gene` (either
            `~pyaragorn.TRNAGene` or `~pyaragorn.TMRNAGene`) corresponding
            to RNA genes detected in the sequence according to the `RNAFinder`
            parameters.

        """
        cdef int      n
        cdef int      nt
        cdef csw      sw
        cdef int*     vsort  = NULL
        cdef Cursor   cursor = Cursor(sequence)

        # copy parameters to ensure the `find_rna` method is re-entrant
        memcpy(&sw, &self._sw, sizeof(csw))

        try:
            with nogil:
                # allocate memory for the result genes
                sw.genespace = aragorn.NT
                sw.genes = <gene*> calloc(sw.genespace, sizeof(gene))
                if sw.genes is NULL:
                    raise MemoryError("failed to allocate memory")
                # detect RNA genes with the "batched" algorithm
                nt = self._bopt(cursor, &sw)
                # allocate array for sorting genes
                vsort = <int*> calloc(nt, sizeof(int))
                if vsort is NULL:
                    raise MemoryError("failed to allocate memory")
                # sort and threshold genes
                n = aragorn.gene_sort(&cursor.ds, nt, vsort, &sw)
            # recover genes
            genes = []
            for i in range(n):
                genes.append(Gene._new_gene(&sw.genes[vsort[i]], sw.geneticcode))
        finally:
            free(vsort)
            free(sw.genes)

        return genes

    cdef int _bopt(
        self,
        Cursor cursor,
        csw* sw
    ) except -1 nogil:
        # adapted from bopt_fastafile to use with our own `Cursor` dataset
        cdef int nt
        cdef int seq[((2 * aragorn.LSEQ) + aragorn.WRAP) + 1]
        cdef int cseq[((2 * aragorn.LSEQ) + aragorn.WRAP) + 1]
        cdef int wseq[(2 * aragorn.WRAP) + 1]
        cdef long i
        cdef long rewind
        cdef long drewind
        cdef long tmaxlen
        cdef bint flag
        cdef int length
        cdef int *s
        cdef int *sf
        cdef int *se
        cdef int *sc
        cdef int *swrap
        cdef long gap
        cdef long start
        cdef bint loop
        cdef bint NX
        cdef bint SH

        # compute width of sliding windows
        rewind = aragorn.MAXTAGDIST + 20
        if sw.trna or sw.mtrna:
            tmaxlen = aragorn.MAXTRNALEN + sw.maxintronlen
            if rewind < tmaxlen:
                rewind = tmaxlen
        if sw.tmrna:
            if rewind < aragorn.MAXTMRNALEN:
                rewind = aragorn.MAXTMRNALEN
        if sw.peptide:
            if sw.tagthresh >= 5 and rewind < aragorn.TSWEEP:
                rewind = aragorn.TSWEEP

        sw.loffset = rewind
        sw.roffset = rewind
        drewind = 2 * rewind

        # cleanly initialize gene array
        aragorn.init_gene(sw.genes, 0, aragorn.NT)

        nt = 0
        flag = 0
        start = 1L

        loop = True
        NX = True
        SH = True

        se = seq
        if sw.linear:
            for i in range(rewind):
                postincrement(se)[0] = aragorn.NOBASE
            start -= rewind
        else:
            if cursor.ds.psmax <= drewind:
                gap = drewind - cursor.ds.psmax
                sc = se + gap
                while se < sc:
                    postincrement(se)[0] = aragorn.NOBASE

                swrap = wseq
                sc = se + cursor.ds.psmax
                while se < sc:
                    se[0] = cursor._forward()
                    postincrement(swrap)[0] = postincrement(se)[0]

                sc = swrap + gap
                while swrap < sc:
                    postincrement(swrap)[0] = aragorn.NOBASE

                swrap = wseq
                sc = swrap + cursor.ds.psmax
                while swrap < sc:
                    postincrement(se)[0] = postincrement(swrap)[0]

                swrap = wseq
                sc = swrap + drewind
                while swrap < sc:
                    postincrement(se)[0] = postincrement(swrap)[0]

                sw.loffset = drewind
                sw.roffset = drewind
                start -= drewind
                flag = 1
                # goto SH
                loop = True
                SH = True
                NX = False

            else:
                swrap = wseq
                sc = seq + drewind
                while se < sc:
                    se[0] = cursor._forward()
                    postincrement(swrap)[0] = postincrement(se)[0]

        # weird ass loop to emulate a GOTO
        while loop:

            # label NX: next
            sc = seq + aragorn.LSEQ
            if NX:
                while (se < sc):
                    postincrement(se)[0] = cursor._forward()
                    if cursor.ds.ps >= cursor.ds.psmax:
                        if sw.linear:
                            for i in range(rewind):
                                postincrement(se)[0] = aragorn.NOBASE
                        else:
                            sc = wseq + drewind
                            swrap = wseq
                            while (swrap < sc):
                                postincrement(se)[0] = postincrement(swrap)[0]
                        flag = 1
                        SH = True
                        break

            # label SH: search
            if SH:
                length = <int> (se - seq)

                with gil:
                    PyErr_CheckSignals()

                # if (sw.verbose):
                #     vstart = sq(d, start + sw.loffset)
                #     vstop = sq(d, ((start + length) - sw.roffset) - 1)
                #     if (vstop < vstart):
                #         fprintf(stderr, "Searching from %ld to %ld\n", vstart, d.psmax)
                #         fprintf(stderr, "Searching from 1 to %ld\n", vstop)
                #     else:
                #         fprintf(stderr, "Searching from %ld to %ld\n", vstart, vstop)

                if (sw.both != 1):
                    sw.start = start
                    sw.comp = 0
                    nt = aragorn.tmioptimise(&cursor.ds, seq, length, nt, sw)

                if (sw.both > 0):
                    aragorn.sense_switch(seq, cseq, length)
                    sw.start = start + length
                    sw.comp = 1
                    nt = aragorn.tmioptimise(&cursor.ds, cseq, length, nt, sw)

                if not flag:
                    s = seq
                    sf = se - drewind
                    se = seq + drewind
                    while (s < se):
                        postincrement(s)[0] = postincrement(sf)[0]
                    start += length - drewind
                    # goto NX
                    NX = SH = loop = True
                    continue

                if nt < 1:
                    cursor.ds.nf += 1
                if sw.maxintronlen > 0:
                    aragorn.remove_overlapping_trna(&cursor.ds, nt, sw)
                if sw.updatetmrnatags:
                    aragorn.update_tmrna_tag_database(sw.genes, nt, sw)

                # FIXME: here should sort genes and filter them with `gene_sort`
                # aragorn.batch_gene_set(d, nt, sw)

                # if sw.verbose:
                #     fprintf(stderr, "%s\nSearch Finished\n\n", d.seqname)

                cursor.ds.ns += 1
                # exit loop
                loop = False

        return nt

    # if (d.ns > 1) and (sw.batch < 2):
    #     fprintf(f, ">end \t%d sequences", d.ns)
    #     if sw.trna or sw.mtrna:
    #         fprintf(f, " %d tRNA genes", sw.ngene[<int> aragorn.tRNA])
    #     if sw.tmrna:
    #         fprintf(f, " %d tmRNA genes", sw.ngene[<int> aragorn.tmRNA])
    #     if d.nf > 0:
    #         sens = (100.0 * (d.ns - d.nf)) / d.ns
    #         fprintf(f, ", nothing found in %d sequences, (%.2lf%% sensitivity)", d.nf, sens)
    #     fputc('\n', f)
    # if sw.updatetmrnatags:
    #     aragorn.report_new_tmrna_tags(sw)



from . import lib
from .lib import (
    __version__,
    Gene,
    TRNAGene,
    TMRNAGene,
    RNAFinder,
    ARAGORN_VERSION,
    TRANSLATION_TABLES,
)

__doc__ = lib.__doc__
__all__ = [
    "Gene",
    "TRNAGene",
    "TMRNAGene",
    "RNAFinder",
    "ARAGORN_VERSION",
    "TRANSLATION_TABLES",
]

__author__ = "Martin Larralde <martin.larralde@embl.de>"
__license__ = "GPL-3.0-or-later"

# Small addition to the docstring: we want to show a link redirecting to the
# rendered version of the documentation, but this can only work when Python
# is running with docstrings enabled
if __doc__ is not None:
    __doc__ += """See Also:
    An online rendered version of the documentation for this version
    of the library on
    `Read The Docs <https://pyaragorn.readthedocs.io/en/v{}/>`_.

    """.format(
        __version__
    )
